from __future__ import annotations

import re
from collections.abc import Iterable, MutableMapping
from typing import Any

from kedro.framework.project import pipelines
from kedro.framework.startup import ProjectMetadata
from kedro.pipeline import Pipeline
from kedro.pipeline.node import Node

from kedro_databricks.cli.bundle.utils import get_entry_point, remove_nulls, sort_dict
from kedro_databricks.constants import (
    MAX_TASK_KEY_LENGTH,
    TASK_KEY_ORDER,
    WORKFLOW_KEY_ORDER,
)
from kedro_databricks.logger import get_logger
from kedro_databricks.utils import make_workflow_name, require_databricks_run_script

log = get_logger("bundle").getChild(__name__)


class ResourceGenerator:
    """Generates Databricks resources for the given pipelines.

    Finds all pipelines in the project and generates Databricks asset bundle resources
    for each according to the Databricks REST API
    """

    def __init__(
        self,
        metadata: ProjectMetadata,
        env: str,
        conf_source: str = "conf",
        params: str | None = None,
    ) -> None:
        self.metadata = metadata
        self.env = env
        self.pipelines: MutableMapping = pipelines
        self.remote_conf_dir = f"/${{workspace.file_path}}/{conf_source}"
        self.params = params

    def generate_resources(
        self, pipeline_name: str | None = None
    ) -> dict[str, dict[str, Any]]:
        """Generate Databricks resources for the given pipelines.

        Finds all pipelines in the project and generates Databricks asset bundle resources
        for each according to the Databricks REST API

        Args:
            pipeline_name (str | None): The name of the pipeline for which Databricks asset bundle resources should be generated.
                If None, generates all pipelines.

        Returns:
            dict[str, dict[str, Any]]: A dictionary of pipeline names and their Databricks resources
        """
        workflows = {}
        pipeline = self.pipelines.get(pipeline_name)
        if pipeline_name and pipeline:
            log.info(f"Generating resources for pipeline '{pipeline_name}'")
            name = make_workflow_name(self.metadata.package_name, pipeline_name)
            workflows[name] = self._create_workflow(name=name, pipeline=pipeline)
            return self._workflows_to_resources(workflows)
        if pipeline_name:
            raise KeyError(
                f"Pipeline '{pipeline_name}' not found. Available pipelines: {list(self.pipelines.keys())}"
            )

        for pipe_name, pipeline in self.pipelines.items():
            if len(pipeline.nodes) == 0:
                continue
            name = make_workflow_name(self.metadata.package_name, pipe_name)
            workflow = self._create_workflow(name=name, pipeline=pipeline)
            log.debug(f"Workflow '{name}' successfully created.")
            log.debug(workflow)
            workflows[name] = workflow

        return self._workflows_to_resources(workflows)

    def _workflows_to_resources(
        self, workflows: dict[str, dict[str, Any]]
    ) -> dict[str, dict[str, Any]]:
        """Convert Databricks workflows to Databricks resources.

        Args:
            workflows (Dict[str, Dict[str, Any]]): A dictionary of Databricks workflows

        Returns:
            Dict[str, Dict[str, Any]]: A dictionary of Databricks resources
        """
        resources = {
            name: {"resources": {"jobs": {name: wf}}} for name, wf in workflows.items()
        }
        log.debug(resources)
        log.info("Databricks resources successfully generated.")
        return resources

    def _create_workflow(self, name: str, pipeline: Pipeline) -> dict[str, Any]:
        """Create a Databricks workflow for a given pipeline.

        Args:
            name (str): name of the pipeline
            pipeline (Pipeline): Kedro pipeline object
            env (str): name of the env to be used by the tasks of the workflow

        Returns:
            Dict[str, Any]: a Databricks workflow
        """
        ## Follows the Databricks REST API schema
        ## https://docs.databricks.com/api/workspace/jobs/create
        workflow = {
            "name": name,
            "tasks": [
                self._create_task(node, depends_on=deps)
                for node, deps in sorted(pipeline.node_dependencies.items())
            ],
        }
        non_null = remove_nulls(sort_dict(workflow, WORKFLOW_KEY_ORDER))
        if not isinstance(non_null, dict):  # pragma: no cover - this is a type check
            raise RuntimeError("Expected a dict")
        return non_null

    def _create_task(
        self,
        node: Node,
        depends_on: Iterable[Node],
    ) -> dict[str, Any]:
        """Create a Databricks task for a given node.

        Args:
            name (str): name of the node
            depends_on (List[Node]): list of nodes that the task depends on
            env (str): name of the env to be used by the task

        Returns:
            Dict[str, Any]: a Databricks task
        """
        ## Follows the Databricks REST API schema. See "tasks" in the link below
        ## https://docs.databricks.com/api/workspace/jobs/create
        entry_point = get_entry_point(self.metadata.project_name)
        params = [
            "--nodes",
            node.name,
            "--conf-source",
            self.remote_conf_dir,
            "--env",
            "${var.environment}",
        ]

        if require_databricks_run_script():  # pragma: no cover
            entry_point = "databricks_run"
            params = params + ["--package-name", self.metadata.package_name]

        if self.params:
            params = params + ["--params", self.params]

        task = {
            "task_key": sanitize_name(node),
            "libraries": [{"whl": "../dist/*.whl"}],
            "depends_on": [
                {"task_key": sanitize_name(dep)}
                for dep in sorted(depends_on, key=lambda dep: dep.name)
            ],
            "python_wheel_task": {
                "package_name": self.metadata.package_name,
                "entry_point": entry_point,
                "parameters": params,
            },
        }

        return sort_dict(task, TASK_KEY_ORDER)


def sanitize_name(node: Node) -> str:
    """Sanitize the node name to be used as a task key in Databricks.

    Args:
        node (Node): Kedro node object

    Returns:
        str: sanitized task key
    """
    _name = node.name
    if not re.match(r"^[\w\-\_]+$", _name):  # Ensure the name is valid
        log.warning(
            f"Node name '{_name}' contains invalid characters and will be sanitized. "
            "To avoid this use an explicit node name as `node(..., name='valid_name')`."
        )

        _name = re.sub(r"[^\w\_]", "_", _name)
        _name = re.sub(r"_{2,}", "_", _name).lstrip("_").rstrip("_")

    if len(_name) > MAX_TASK_KEY_LENGTH:  # Ensure the name is not too long
        log.warning(
            f"Node name '{_name}' is too long. "
            f"Truncating to {MAX_TASK_KEY_LENGTH} characters."
        )
        _name = _name[:MAX_TASK_KEY_LENGTH]

    return _name



from __future__ import annotations

from collections import namedtuple
from collections.abc import MutableMapping

from databricks.sdk import WorkspaceClient
from databricks.sdk.service.jobs import BaseJob
from kedro.framework.project import pipelines as kedro_pipelines
from kedro.framework.startup import ProjectMetadata

from kedro_databricks.logger import get_logger
from kedro_databricks.utils import make_workflow_name

JobLink = namedtuple("JobLink", ["name", "url", "is_dev"])


log = get_logger("deploy").getChild("get_deployed_resources")


# TODO: Add tests
def get_deployed_resources(
    metadata: ProjectMetadata,
    pipelines: MutableMapping = kedro_pipelines,
    only_dev: bool = False,
    _custom_username: str | None = None,
) -> set[JobLink]:
    """Print deployed pipelines.

    Args:
        pipelines (_ProjectPipelines): Project pipelines.
        only_dev (bool): Whether to show only dev jobs.
        _custom_username (str): Custom username to use.

    Returns:
        dict[str, set[str]]: Deployed pipelines.
    """
    w = WorkspaceClient()
    job_host = f"{w.config.host}/jobs"
    username = _get_username(_custom_username or w.current_user.me().user_name)
    log.info(f"Getting jobs for {username}")
    all_jobs = {
        job.settings.name: job
        for job in w.jobs.list()
        if job.settings is not None and job.settings.name is not None
    }
    jobs = _gather_user_jobs(metadata, all_jobs, pipelines, username, job_host)
    for job in jobs:
        if only_dev and not job.is_dev:  # pragma: no cover
            continue
        log.info(f"Run '{job.name}' at {job.url}")
    return jobs


# TODO: Add tests
def _gather_user_jobs(
    metadata: ProjectMetadata,
    all_jobs: dict[str, BaseJob],
    pipelines: MutableMapping,
    username: str,
    job_host: str,
) -> set[JobLink]:
    jobs = set()
    for job_name, job in all_jobs.items():
        is_dev = job_name.startswith("[dev")
        is_valid = _is_valid_job(metadata, pipelines, job_name)
        if (is_dev and username not in job_name) or not is_valid:  # pragma: no cover
            continue
        n = job_name.split(" - ")[0]
        link = JobLink(name=n, url=f"{job_host}/{job.job_id}", is_dev=is_dev)
        jobs.add(link)
    return jobs


def _is_valid_job(
    metadata: ProjectMetadata, pipelines: MutableMapping, job_name: str
) -> bool:
    return any(
        make_workflow_name(metadata.package_name, pipeline_name) in job_name
        for pipeline_name in pipelines
    )


def _get_username(username: str | None):
    if username is None:  # pragma: no cover
        raise ValueError("Could not get username from Databricks")
    if "@" in username:
        username = username.split("@")[0]
    return username



from __future__ import annotations

import click
from kedro.framework.cli.project import CONF_SOURCE_HELP, PIPELINE_ARG_HELP
from kedro.framework.cli.utils import ENV_HELP
from kedro.framework.startup import ProjectMetadata

import kedro_databricks
from kedro_databricks import cli
from kedro_databricks.constants import (
    DEFAULT_CONF_FOLDER,
    DEFAULT_CONFIG_HELP,
    DEFAULT_CONFIG_KEY,
    DEFAULT_PROVIDER,
    DEFAULT_TARGET,
    PROVIDER_PROMPT,
)


@click.group(name="Kedro-Databricks")
def commands():
    """Entry point for Kedro-Databricks commands"""
    pass


@commands.group(name="databricks")
def databricks_commands():
    """Databricks Asset Bundle commands

    These commands are used to manage Databricks Asset Bundles in a Kedro project.
    They allow you to initialize, bundle, deploy, run, and destroy Databricks asset bundles.
    """
    pass


@databricks_commands.command(context_settings=dict(ignore_unknown_options=True))
def version():
    """Display the version of Kedro-Databricks

    This command prints the version of the Kedro-Databricks plugin.
    It is useful for checking the installed version of the plugin in your Kedro project.
    """
    print(f"Kedro-Databricks version: {kedro_databricks.__version__}")  # noqa: T201


@databricks_commands.command(context_settings=dict(ignore_unknown_options=True))
@click.option(
    "-d", "--default-key", default=DEFAULT_CONFIG_KEY, help=DEFAULT_CONFIG_HELP
)
@click.option("--provider", prompt=PROVIDER_PROMPT, default=DEFAULT_PROVIDER)
@click.argument("databricks_args", nargs=-1, type=click.UNPROCESSED)
@click.pass_obj
def init(
    metadata: ProjectMetadata,
    default_key: str,
    provider: str,
    databricks_args: list[str],
):
    """Initialize Databricks Asset Bundle configuration

    This command initializes the Databricks Asset Bundle configuration in the Kedro project.
    The `default_key` is the configuration key to use for Databricks, and the `provider`
    specifies the cloud provider for the Databricks Asset Bundle (e.g., "aws", "azure").

    This command will create a `databricks.yml` file in the `conf/<env>` directory
    and set up the necessary configuration for the Databricks Asset Bundle.

    To see additional options, run `databricks bundle init --help`.

    Args:
        metadata: Project metadata containing project path and other information.
        default_key: The default configuration key to use for Databricks.
        provider: The provider for the Databricks Asset Bundle (e.g., "aws", "azure").
        databricks_args: Additional arguments to pass to the `databricks` CLI.
    """
    cli.init(metadata, provider, default_key, *databricks_args)


@databricks_commands.command()
@click.option(
    "-d", "--default-key", default=DEFAULT_CONFIG_KEY, help=DEFAULT_CONFIG_HELP
)
@click.option("-e", "--env", default=DEFAULT_TARGET, help=ENV_HELP)
@click.option("-c", "--conf-source", default=DEFAULT_CONF_FOLDER, help=CONF_SOURCE_HELP)
@click.option("-p", "--pipeline", default=None, help=PIPELINE_ARG_HELP)
@click.option(
    "-r",
    "--params",
    default=None,
    help="Kedro run time params in `key1=value1,key2=value2` format",
)
@click.option(
    "--overwrite",
    default=False,
    is_flag=True,
    show_default=True,
    help="Overwrite the existing resources",
)
@click.pass_obj
def bundle(
    metadata: ProjectMetadata,
    default_key: str,
    env: str,
    conf_source: str,
    pipeline: str | None,
    params: str | None,
    overwrite: bool,
):
    """Convert kedro pipelines into Databricks asset bundle resources

    This command generates asset bundle resources for the specified Kedro pipeline
    and saves them in the `resources` directory.
    To validate the generated resources, you can run the `databricks bundle validate` command.

    Args:
        metadata: Project metadata containing project path and other information.
        default_key: The default configuration key to use for Databricks.
        env: The environment for the Kedro project (e.g., "dev", "prod").
        conf_source: The source of the Kedro configuration files.
        pipeline: The pipeline to bundle (optional).
        params: Kedro run time parameters in `key1=value1,key2=value2` format (optional).
        overwrite: Whether to overwrite existing resources.
    """
    cli.bundle(
        metadata=metadata,
        env=env,
        conf_source=conf_source,
        pipeline_name=pipeline,
        default_key=default_key,
        params=params,
        overwrite=overwrite,
    )


@databricks_commands.command(context_settings=dict(ignore_unknown_options=True))
@click.option("-e", "--env", default=DEFAULT_TARGET, help=ENV_HELP)
@click.option(
    "-b",
    "--bundle/--no-bundle",
    default=False,
    help="Bundle the project before deploying",
)
@click.option("-c", "--conf-source", default=DEFAULT_CONF_FOLDER, help=CONF_SOURCE_HELP)
@click.option("-p", "--pipeline", default=None, help=PIPELINE_ARG_HELP)
@click.option(
    "-r",
    "--runtime-params",
    default=None,
    help="Kedro run time params in `key1=value1,key2=value2` format",
)
@click.argument("databricks_args", nargs=-1, type=click.UNPROCESSED)
@click.pass_obj
def deploy(
    metadata: ProjectMetadata,
    env: str,
    bundle: bool,
    conf_source: str,
    pipeline: str,
    runtime_params: str | None,
    databricks_args: list[str],
):
    """Deploy the asset bundle to Databricks

    This command deploys the Databricks asset bundle to the specified environment.
    If `--bundle` is specified, it will first bundle the project into resources.
    To see additional options, run `databricks bundle deploy --help`.

    Args:
        metadata: Project metadata containing project path and other information.
        env: The environment for the Kedro project (e.g., "dev", "prod").
        bundle: Whether to bundle the project before deploying.
        conf_source: The source of the Kedro configuration files.
        pipeline: The pipeline to deploy (optional).
        runtime_params: Kedro run time parameters in `key1=value1,key2=value2` format (optional).
        databricks_args: Additional arguments to pass to the `databricks` CLI.
    """
    if bundle:
        cli.bundle(
            metadata=metadata,
            env=env,
            pipeline_name=pipeline,
            conf_source=conf_source,
            default_key=DEFAULT_CONFIG_KEY,
            params=runtime_params,
            overwrite=True,
        )
    cli.deploy(metadata, env, *databricks_args)


@databricks_commands.command(context_settings=dict(ignore_unknown_options=True))
@click.argument("pipeline")
@click.argument("databricks_args", nargs=-1, type=click.UNPROCESSED)
@click.pass_obj
def run(metadata: ProjectMetadata, pipeline: str, databricks_args: list[str]):
    """Run the project on Databricks

    This is a wrapper for the `databricks bundle run` command.
    To see additional options, run `databricks bundle run --help`.

    Args:
        metadata: Project metadata containing project path and other information.
        pipeline: The pipeline to run on Databricks.
        databricks_args: Additional arguments to be passed to the `databricks` CLI.
    """
    cli.run(metadata, pipeline, *databricks_args)


@databricks_commands.command(context_settings=dict(ignore_unknown_options=True))
@click.option("-e", "--env", default=DEFAULT_TARGET, help=ENV_HELP)
@click.argument("databricks_args", nargs=-1, type=click.UNPROCESSED)
@click.pass_obj
def destroy(metadata: ProjectMetadata, env: str, databricks_args: list[str]):
    """Destroy the Databricks asset bundle

    This is a wrapper for the `databricks bundle destroy` command.
    To see additional options, run `databricks bundle destroy --help`.

    Args:
        metadata: Project metadata containing project path and other information.
        env: The environment for the Kedro project (e.g., "dev", "prod").
        databricks_args: Additional arguments to be passed to the `databricks` CLI.
    """
    cli.destroy(metadata, env, *databricks_args)



import shutil
from collections.abc import Generator

import pytest
import yaml
from click.testing import CliRunner
from kedro.framework.startup import ProjectMetadata

from kedro_databricks.cli.deploy import _get_arg_value
from kedro_databricks.constants import DEFAULT_TARGET
from kedro_databricks.plugin import commands
from kedro_databricks.utils import get_targets, read_databricks_config
from tests.utils import reset_init


def _load_resources(metadata, required_files):
    resource_dir = metadata.project_path / "resources"
    assert resource_dir.exists(), "Resource directory not created"
    assert resource_dir.is_dir(), "Resource directory is not a directory"
    files = [p.name for p in resource_dir.rglob("*")]
    files.sort()
    assert files == required_files, f"Resource files not created: {', '.join(files)}"
    resources = []
    for p in files:
        with open(resource_dir / p) as f:
            resource = yaml.safe_load(f)
            resources.append(resource)
    return zip(resources, files)


def _reset_bundle(metadata):
    """Reset the bundle to its initial state."""
    bundle_dir = metadata.project_path / "resources"
    if bundle_dir.exists():
        shutil.rmtree(bundle_dir)


def _validate_resources(metadata, required_files, task_validator):
    """Validate the resources generated by the bundle command."""
    resources = _load_resources(
        metadata=metadata,
        required_files=required_files,
    )
    for resource, file_name in resources:
        assert resource.get("resources") is not None

        jobs = resource["resources"]["jobs"]
        assert len(jobs) == 1

        job = jobs.get(file_name.split(".")[0])
        assert job is not None

        tasks = job.get("tasks")
        assert tasks is not None
        task_validator(tasks)


@pytest.fixture(scope="module")
def init_kedro_project(
    cli_runner, metadata, custom_provider
) -> Generator[tuple[ProjectMetadata, CliRunner]]:
    reset_init(metadata)
    _reset_bundle(metadata)
    init_cmd = ["databricks", "init", "--provider", custom_provider]
    result = cli_runner.invoke(commands, init_cmd, obj=metadata)
    assert result.exit_code == 0, (result.exit_code, result.stdout, result.exception)
    assert metadata.project_path.exists(), "Project path not created"
    assert metadata.project_path.is_dir(), "Project path is not a directory"
    assert metadata.project_path / "databricks.yml", "Databricks config not created"

    databricks_config = read_databricks_config(metadata.project_path)
    targets = get_targets(databricks_config)
    for target in targets:
        override_path = metadata.project_path / "conf" / target / "databricks.yml"
        assert (
            override_path.exists()
        ), f"Resource Overrides at {override_path} does not exist"
    yield metadata, cli_runner
    reset_init(metadata)
    _reset_bundle(metadata)


def test_databricks_bundle_fail(cli_runner, metadata):
    bundle_fail = ["databricks", "bundle", "--default-key", "_deault"]
    result = cli_runner.invoke(commands, bundle_fail, obj=metadata)
    assert result.exit_code == 1, (result.exit_code, result.stdout, result.exception)


def test_databricks_bundle_with_overrides(init_kedro_project):
    metadata, cli_runner = init_kedro_project

    command = ["databricks", "bundle", "--env", DEFAULT_TARGET]
    result = cli_runner.invoke(commands, command, obj=metadata)
    assert result.exit_code == 0, (result.exit_code, result.stdout, result.exception)

    def task_validator(tasks):
        assert len(tasks) == 8
        for i, task in enumerate(tasks):
            assert task.get("task_key") in (f"node{i}", f"ns_{i}_node_{i}_1")
            assert task.get("job_cluster_key") == "default"
            params = task.get("python_wheel_task").get("parameters")
            assert params is not None
            env = _get_arg_value(params, "--env")
            assert env is not None
            assert env == "${var.environment}"

    _validate_resources(
        metadata=metadata,
        required_files=[
            f"{metadata.package_name}.yml",
            f"{metadata.package_name}_ds.yml",
            f"{metadata.package_name}_namespaced_pipeline.yml",
        ],
        task_validator=task_validator,
    )


def test_databricks_bundle_with_conf(init_kedro_project):
    """Test the `bundle` command"""
    metadata, cli_runner = init_kedro_project

    command = [
        "databricks",
        "bundle",
        "--env",
        DEFAULT_TARGET,
        "--conf-source",
        "conf/sub_pipeline",
    ]
    result = cli_runner.invoke(commands, command, obj=metadata)
    assert result.exit_code == 0, (result.exit_code, result.stdout, result.exception)

    def task_validator(tasks):
        assert len(tasks) == 8
        for i, task in enumerate(tasks):
            assert task.get("task_key") in (f"node{i}", f"ns_{i}_node_{i}_1")
            assert task.get("job_cluster_key") == "default"
            params = task.get("python_wheel_task").get("parameters")
            assert params is not None
            env = _get_arg_value(params, "--env")
            assert env is not None
            assert env == "${var.environment}"

    _validate_resources(
        metadata=metadata,
        required_files=[
            f"{metadata.package_name}.yml",
            f"{metadata.package_name}_ds.yml",
            f"{metadata.package_name}_namespaced_pipeline.yml",
        ],
        task_validator=task_validator,
    )


def test_databricks_bundle_without_overrides(init_kedro_project):
    """Test the `bundle` command"""
    metadata, cli_runner = init_kedro_project

    command = ["databricks", "bundle", "--env", DEFAULT_TARGET]
    result = cli_runner.invoke(commands, command, obj=metadata)
    assert result.exit_code == 0, (result.exit_code, result.stdout, result.exception)

    def task_validator(tasks):
        assert len(tasks) == 8
        for i, task in enumerate(tasks):
            assert task.get("task_key") in (f"node{i}", f"ns_{i}_node_{i}_1")
            assert task.get("job_cluster_key") == "default"
            params = task.get("python_wheel_task").get("parameters")
            assert params is not None
            env = _get_arg_value(params, "--env")
            assert env is not None
            assert env == "${var.environment}"

    _validate_resources(
        metadata=metadata,
        required_files=[
            f"{metadata.package_name}.yml",
            f"{metadata.package_name}_ds.yml",
            f"{metadata.package_name}_namespaced_pipeline.yml",
        ],
        task_validator=task_validator,
    )


def test_databricks_bundle_with_params(init_kedro_project):
    """Test the `bundle` command"""
    metadata, cli_runner = init_kedro_project
    command = [
        "databricks",
        "bundle",
        "--env",
        DEFAULT_TARGET,
        "--params",
        "run_date={{job.parameters.run_date}},run_id={{job.parameters.run_id}}",
        "--overwrite",
    ]
    result = cli_runner.invoke(commands, command, obj=metadata)
    assert result.exit_code == 0, (result.exit_code, result.stdout, result.exception)

    def task_validator(tasks):
        assert len(tasks) == 8
        for i, task in enumerate(tasks):
            assert task.get("task_key") in (f"node{i}", f"ns_{i}_node_{i}_1")
            assert task.get("job_cluster_key") == "default"
            params = task.get("python_wheel_task").get("parameters")
            assert params is not None
            value = _get_arg_value(params, "--params")
            assert value is not None
            assert (
                value
                == "run_date={{job.parameters.run_date}},run_id={{job.parameters.run_id}}"
            )

    _validate_resources(
        metadata=metadata,
        required_files=[
            f"{metadata.package_name}.yml",
            f"{metadata.package_name}_ds.yml",
            f"{metadata.package_name}_namespaced_pipeline.yml",
        ],
        task_validator=task_validator,
    )


def test_databricks_bundle_with_pipeline(init_kedro_project):
    """Test the `bundle` command"""
    metadata, cli_runner = init_kedro_project
    if (metadata.project_path / "resources").exists():
        shutil.rmtree(metadata.project_path / "resources")
    command = [
        "databricks",
        "bundle",
        "--env",
        DEFAULT_TARGET,
        "--pipeline",
        "ds",
        "--overwrite",
    ]
    result = cli_runner.invoke(commands, command, obj=metadata)
    assert result.exit_code == 0, (result.exit_code, result.stdout, result.exception)

    def task_validator(tasks):
        assert len(tasks) == 8
        for i, task in enumerate(tasks):
            assert task.get("task_key") in (f"node{i}", f"ns_{i}_node_{i}_1")
            assert task.get("job_cluster_key") == "default"
            params = task.get("python_wheel_task").get("parameters")
            assert params is not None
            env = _get_arg_value(params, "--env")
            assert env is not None
            assert env == "${var.environment}"
            pipeline = _get_arg_value(params, "--pipeline")
            assert pipeline is None

    _validate_resources(
        metadata=metadata,
        required_files=[
            f"{metadata.package_name}_ds.yml",
        ],
        task_validator=task_validator,
    )


def test_databricks_bundle_with_no_nodes(init_kedro_project):
    """Test the `bundle` command"""
    metadata, cli_runner = init_kedro_project
    command = [
        "databricks",
        "bundle",
        "--env",
        DEFAULT_TARGET,
        "--pipeline",
        "non-existing-pipeline",
        "--overwrite",
    ]
    result = cli_runner.invoke(commands, command, obj=metadata)
    assert result.exit_code == 1, (result.exit_code, result.stdout, result.exception)
    assert isinstance(result.exception, KeyError)
    exception = bytes(str(result.exception), "utf-8").decode("unicode_escape")
    expected = "Pipeline 'non-existing-pipeline' not found."
    assert (
        expected in result.exception.args[0]
    ), f"Expected exception message: {expected} not found in {exception}"



from __future__ import annotations

import logging
import re
import shutil
import subprocess
from pathlib import Path

import yaml
from packaging.version import Version

from kedro_databricks.constants import KEDRO_VERSION, MINIMUM_DATABRICKS_VERSION
from kedro_databricks.logger import get_logger

log = get_logger("utils")


class Command:
    """A class to run shell commands and handle their output.

    This class provides a way to run shell commands while capturing their output
    and logging it. It can also handle errors by either raising an exception or
    logging a warning, depending on the `warn` parameter.

    Attributes:
        command (list[str]): The command to run as a list of strings.
        log (logging.Logger): The logger to use for logging output.
        warn (bool): If True, log a warning instead of raising an exception on error.

    Args:
        command (list[str]): The command to run as a list of strings.
        log (logging.Logger): The logger to use for logging output.
        warn (bool): If True, log a warning instead of raising an exception on error.
    """

    def __init__(self, command: list[str], log: logging.Logger, warn: bool = False):
        self.log = log.getChild("command")
        self.command = command
        self.warn = warn

    def __str__(self):  # pragma: no cover
        return f"Command({self.command})"

    def __repr__(self):  # pragma: no cover
        return self.__str__()

    def __rich_repr__(self):  # pragma: no cover
        yield "program", self.command[0]
        yield "args", self.command[1:]

    def _read_stdout(self, process: subprocess.Popen):
        stdout = []
        while True:
            line = process.stdout.readline()  # type: ignore - we know it's there
            if not line and process.poll() is not None:
                break
            print(line, end="")  # noqa: T201
            stdout.append(line)
        return stdout

    def _run_command(self, command, **kwargs):
        """Run a command while printing the live output"""
        process = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
            **kwargs,
        )
        stdout = self._read_stdout(process)
        process.stdout.close()  # type: ignore - we know it's there
        result = subprocess.CompletedProcess(
            args=command,
            returncode=process.returncode,
            stdout=stdout or [""],
            stderr=[],
        )
        if result.returncode != 0 and "deploy" not in command:
            self._handle_error(result)
        return result

    def run(self, *args, **kwargs):
        cmd = self.command + list(*args)
        self.log.info(f"Running command: {cmd}")
        return self._run_command(cmd, **kwargs)

    def _handle_error(self, result: subprocess.CompletedProcess):
        stdout = "\n".join(result.stdout)
        error_msg = f"Failed to run command `{' '.join(self.command)}`:\n\n{stdout}"
        if self.warn:
            self.log.warning(error_msg)
        else:
            raise RuntimeError(error_msg)


def make_workflow_name(package_name, pipeline_name: str) -> str:
    """Create a name for the Databricks workflow.

    Args:
        pipeline_name (str): The name of the pipeline

    Returns:
        str: The name of the workflow
    """
    if pipeline_name == "__default__":
        return package_name
    sanitised_pipeline_name = pipeline_name.replace(".", "_")
    return f"{package_name}_{sanitised_pipeline_name}"


def read_databricks_config(project_path: Path) -> dict:
    """Read the databricks.yml configuration file.

    Args:
        project_path (Path): The path to the Kedro project.

    Returns:
        dict: The configuration as a dictionary.
    """
    with open(project_path / "databricks.yml") as f:
        conf = yaml.safe_load(f)
    return conf


def get_bundle_name(config: dict) -> str:
    """Get the bundle name from the databricks.yml configuration.

    Args:
        config (dict): The configuration as a dictionary.

    Returns:
        str: The bundle name.

    Raises:
        ValueError: If the bundle name is not found.
    """
    bundle_name = config.get("bundle", {}).get("name")
    if bundle_name is None:
        raise ValueError("No bundle name found in databricks.yml")
    return bundle_name


def get_targets(config: dict) -> dict:
    """Get the targets from the databricks.yml configuration.

    Args:
        config (dict): The configuration as a dictionary.

    Returns:
        dict: The targets as a dictionary.

    Raises:
        ValueError: If the targets are not found.
    """
    targets = config.get("targets")
    if targets is None:
        raise ValueError("No targets found in databricks.yml")
    return targets


def assert_databricks_cli(raise_error: bool = True):  # pragma: no cover
    """Check if the Databricks CLI is installed.

    Raises:
        RuntimeError: If the Databricks CLI is not installed or the wrong version is used.
        RuntimeError: If the Databricks CLI version is less than the minimum required version.
    """
    if shutil.which("databricks") is None:  # pragma: no cover
        msg = "Databricks CLI is not installed."
        if raise_error:
            raise RuntimeError(msg)
        return msg
    return _check_version(raise_error)


def require_databricks_run_script(_version=KEDRO_VERSION) -> bool:
    """Check if the current Kedro version is less than 0.19.8.

    Kedro 0.19.8 introduced a new `run_script` method that is required for
    running tasks on Databricks. This method is not available in earlier
    versions of Kedro. This function checks if the current Kedro version is
    less than 0.19.8.

    Returns:
        bool: whether the current Kedro version is less than 0.19.8
    """
    return _version < Version("0.19.8")


def _check_version(raise_error: bool = True):
    current_databricks_version = _get_databricks_cli_version()
    if current_databricks_version < MINIMUM_DATABRICKS_VERSION:
        error_msg = f"""{_version_to_str(current_databricks_version)} < {_version_to_str(MINIMUM_DATABRICKS_VERSION)}
    Your Databricks CLI version is {_version_to_str(current_databricks_version)},
    but this script requires at least {_version_to_str(MINIMUM_DATABRICKS_VERSION)}.
    Visit https://docs.databricks.com/en/dev-tools/cli/install.html to install the latest version.
        """
        if raise_error:
            raise RuntimeError(error_msg)
        return error_msg


def _get_databricks_cli_version() -> list[int]:
    result = Command(["databricks", "--version"], log, warn=True).run()
    stdout = "\n".join(result.stdout).strip()
    version_str = re.sub(r".*(\d+\.\d+\.\d+)", r"\1", stdout)
    return list(map(int, version_str.split(".")))


def _version_to_str(version: list[int]) -> str:
    if len(version) != 3:  # noqa: PLR2004 - Semantic versioning requires 3 parts
        raise ValueError(f"Invalid version: {version}")
    return ".".join(str(x) for x in version)



from dataclasses import dataclass
from pathlib import Path

import pytest
import yaml

from kedro_databricks.cli.bundle.override_resources import override_resources

EXAMPLE_ROOT = Path(__file__).parent.parent.parent / "examples"
EXAMPLES = [p.name for p in EXAMPLE_ROOT.iterdir() if p.is_dir()]


@dataclass
class Example:
    overrides: dict
    resources: dict
    result: dict


def _load_example(name):
    path = EXAMPLE_ROOT / name
    with open(path / "databricks.yml") as f:
        overrides = yaml.safe_load(f)
    with open(path / "resources.yml") as f:
        resources = yaml.safe_load(f)
    with open(path / "result.yml") as f:
        result = yaml.safe_load(f)
    return Example(overrides, resources, result)


@pytest.mark.parametrize("example_name", EXAMPLES)
def test_examples(example_name):
    example = _load_example(example_name)
    result = override_resources(example.resources, example.overrides, "default")
    assert result == example.result



import argparse
import logging
import shutil
import subprocess
import sys
from pathlib import Path

import tomlkit

root = Path(__file__).parent.parent
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(levelname)s - %(message)s",
)
log = logging.getLogger(__name__)


def main():
    args = parse_args(sys.argv[1:])
    match args.command:
        case "sync":
            sync(Path(args.name))
        case "new":
            new(args.name, root / args.name, args.overwrite)
        case _:
            log.error(f"Unknown command: {args.command}")
            sys.exit(1)


def current_version() -> str:
    """Get the current version of the project."""
    pyproject = root / "pyproject.toml"

    if not pyproject.exists():
        raise FileNotFoundError(f"pyproject.toml not found at {pyproject}.")

    with open(pyproject, "rb") as f:
        project = tomlkit.load(f)

    return project.get("project", {}).get("version", "0.0.0")


def parse_args(args: list[str]):
    parser = argparse.ArgumentParser(description="Create a new development project")
    subparsers = parser.add_subparsers(dest="command")

    sync_parser = subparsers.add_parser("sync")

    for obj in [sync_parser]:
        obj.add_argument(
            "name", type=str, default="develop-eggs", help="Name of the project"
        )

    new_parser = subparsers.add_parser("new")
    for obj in [new_parser]:
        obj.add_argument(
            "name", type=str, default="develop-eggs", help="Name of the project"
        )
        obj.add_argument(
            "-o",
            "--overwrite",
            action="store_true",
            help="Overwrite the project if it already exists",
            default=False,
        )

    return parser.parse_args()


def new(project_name: str, project_path: Path, overwrite: bool = False):
    if project_path.exists() and not overwrite:
        log.error(f"Project '{project_name}' already exists at {project_path}.")
    elif project_path.exists() and overwrite:
        log.warning(
            f"Project '{project_name}' already exists at {project_path}. Overwriting."
        )
        shutil.rmtree(project_path)
    log.info(f"Created project '{project_name}' at {project_path}.")
    subprocess.run(
        [
            "uv",
            "run",
            "kedro",
            "new",
            "--starter",
            "databricks-iris",
            "--name",
            project_name,
        ],
        check=False,
    )

    with open(project_path / "pyproject.toml", "rb") as f:
        project = tomlkit.load(f)

    project["project"]["requires-python"] = ">=3.10"  # type: ignore

    with open(project_path / "pyproject.toml", "w") as f:
        tomlkit.dump(project, f)

    Path(project_path / ".tool-versions").write_text("uv 0.7.8\n")
    Path(project_path / ".python-version").write_text("3.11\n")
    sync(project_path)


def sync(project_path: Path):
    """Synchronize the development environment."""
    whl_name = f"kedro_databricks-{current_version()}-py3-none-any.whl"
    whl_src = root / "dist" / whl_name
    whl_dst = project_path / whl_name
    log.info(f"Synchronizing development environment for project at {project_path}.")
    subprocess.run(["uv", "build"], check=False, cwd=root)
    shutil.copy(whl_src, whl_dst)
    subprocess.run(["uv", "add", whl_name, "--upgrade"], check=False, cwd=project_path)
    log.info("Development environment synchronized.")


if __name__ == "__main__":
    main()



import subprocess
import tempfile
from pathlib import Path

import mkdocs_gen_files

mkdocs_gen_files.log.info("Generating changelog...")


tmp = Path(tempfile.mktemp()).with_suffix(".md")
subprocess.run(["uv", "run", "cz", "ch", "--file-name", tmp.as_posix()], check=False)
with mkdocs_gen_files.open("changelog.md", "w") as f:
    f.write(tmp.read_text())
mkdocs_gen_files.log.info("Changelog generated successfully.")

tmp.unlink(missing_ok=True)  # Clean up the temporary file
mkdocs_gen_files.log.info("Temporary file cleaned up.")



from __future__ import annotations

import os
from pathlib import Path
from subprocess import CompletedProcess

from kedro.framework.startup import ProjectMetadata

from kedro_databricks.cli.deploy.get_deployed_resources import get_deployed_resources
from kedro_databricks.logger import get_logger
from kedro_databricks.utils import Command, assert_databricks_cli

log = get_logger("deploy")


def deploy(metadata: ProjectMetadata, env: str, *databricks_args: str):
    """Deploy the Databricks Asset Bundle.

    This function deploys the Databricks Asset Bundle in the current project
    directory. It also creates a Databricks configuration file and a
    Databricks target configuration file.

    Args:
        metadata (ProjectMetadata): The project metadata.
        env (str): The environment to deploy to.
        *databricks_args: Additional arguments to be passed to the `databricks` CLI.

    Raises:
        RuntimeError: If the `databricks` CLI is not installed or the wrong version is used.
    """
    _validate_project(metadata)

    os.chdir(metadata.project_path)
    _build_project(metadata)
    result = _upload_project_data(metadata, env)
    if result and result.returncode != 0:  # pragma: no cover
        raise RuntimeError("Failed to upload project data to DBFS")
    _deploy_project(metadata, env, list(databricks_args))


def _validate_project(metadata: ProjectMetadata):
    assert_databricks_cli()
    if not metadata.project_path.exists():
        raise FileNotFoundError(f"Project path {metadata.project_path} does not exist")
    if not (Path(metadata.project_path) / "databricks.yml").exists():
        raise FileNotFoundError(
            "Databricks configuration file does not exist. "
            "Please run `kedro databricks init` to create it."
        )


def _build_project(metadata: ProjectMetadata):
    """Build the project."""
    result = Command(["kedro", "package"], log=log).run(cwd=metadata.project_path)
    return result


# TODO: Add tests
def _upload_project_data(metadata: ProjectMetadata, env: str):  # pragma: no cover
    """Upload the project data to DBFS."""
    source_path = metadata.project_path / "data"
    target_path = f"dbfs:/FileStore/{metadata.package_name}/{env}/data"
    if not source_path.exists():
        log.warning(f"Data path {source_path} does not exist")
        return

    log.info(
        f"Uploading {source_path.relative_to(metadata.project_path)} to {target_path}"
    )
    cmd = [
        "databricks",
        "fs",
        "cp",
        "-r",
        "--overwrite",
        source_path.as_posix(),
        target_path,
    ]
    result = Command(cmd, log=log).run(cwd=metadata.project_path)
    log.info(f"Data uploaded to {target_path}")
    return result


# TODO: Add tests
def _deploy_project(
    metadata: ProjectMetadata, env: str, databricks_args: list[str]
):  # pragma: no cover
    """Deploy the project to Databricks.

    Args:
        databricks_args (list[str]): Databricks arguments.

    Returns:
        subprocess.CompletedProcess: The result of the deployment.
    """
    deploy_cmd = ["databricks", "bundle", "deploy"] + databricks_args
    target = _get_arg_value(databricks_args, "--target")
    if target is None:
        deploy_cmd += ["--target", env]
    result = Command(deploy_cmd, log=log, warn=True).run(cwd=metadata.project_path)
    # databricks bundle deploy logs to stderr for some reason.
    if _check_deployment_complete(result):
        result.returncode = 0
    log.info("Successfully Deployed Jobs")
    get_deployed_resources(metadata, only_dev=target in ["dev", "local"])
    return result


def _get_arg_value(args: list[str], arg_name: str) -> str | None:
    for i, arg in enumerate(args):
        if "=" in arg:
            _arg, value = arg.split("=", 1)
            if _arg == arg_name:
                return value
        elif arg == arg_name:
            return args[i + 1]


def _check_deployment_complete(result: CompletedProcess) -> bool:
    complete_stdout = result.stdout and "Deployment complete!" in result.stdout
    complete_stderr = result.stderr and "Deployment complete!" in result.stderr
    return complete_stdout is True or complete_stderr is True



from importlib import metadata, resources

from packaging.version import Version

TEMPLATES = resources.files("kedro_databricks").joinpath("templates")

KEDRO_VERSION = Version(metadata.version("kedro"))
"""Kedro version used to build this plugin."""

MINIMUM_DATABRICKS_VERSION = [0, 205, 0]
"""Minimum Databricks version required for this plugin."""

MAX_TASK_KEY_LENGTH = 100
"""Maximum number of characters in a task key in Databricks jobs."""

TASK_KEY_ORDER = [
    "task_key",
    "job_cluster_key",
    "new_cluster",
    "depends_on",
    "spark_python_task",
    "python_wheel_task",
]
"""Order of keys in the task configuration for Databricks jobs."""

WORKFLOW_KEY_ORDER = [
    "name",
    "tags",
    "access_control_list",
    "email_notifications",
    "schedule",
    "max_concurrent_runs",
    "job_clusters",
    "tasks",
]
"""Order of keys in the workflow configuration for Databricks jobs."""

DEFAULT_TARGET = "dev"
"""Default target environment for Databricks configurations."""

DEFAULT_CONF_FOLDER = "conf"
"""Default folder for Kedro configurations."""

DEFAULT_CONFIG_KEY = "default"
"""Default configuration key for Databricks jobs."""

DEFAULT_CONFIG_HELP = "Set the key for the default configuration"
"""Help text for the default configuration key option."""

DEFAULT_PROVIDER = "azure"
"""Default cloud provider for Databricks Asset Bundle."""

OVERRIDE_KEY_MAP = {
    "job_clusters": "job_cluster_key",
    "tasks": "task_key",
    "depends_on": "task_key",
    "environments": "environment_key",
    "on_start": "id",
    "on_success": "id",
    "on_failure": "id",
    "on_duration_warning_threshold_exceeded": "id",
    "on_streaming_backlog_exceeded": "id",
    "rules": "metric",
    "parameters": "name",
    "libraries": "whl",
    "subscriptions": "user_name",
}
"""Map of keys that should be overridden in the Databricks configuration."""

IGNORED_OVERRIDE_KEYS = ["init_scripts", "access_control_list"]
"""Keys that should be ignored when overriding the Databricks configuration."""

NODE_TYPE_MAP = {
    "aws": "i3.xlarge",
    "azure": "Standard_DS3_v2",
    "gcp": "n1-standard-4",
}
"""Map of node types for different cloud providers."""

PROVIDER_PROMPT = """
Please select your cloud provider:
- azure
- aws
- gcp
"""
"""Prompt for selecting the cloud provider for Databricks Asset Bundle."""

INVALID_CONFIG_MSG = """
No `databricks.yml` file found. Maybe you forgot to initialize the Databricks bundle?

You can initialize the Databricks bundle by running:

```
kedro databricks init
```
"""
"""Message displayed when no `databricks.yml` file is found in the project."""

GITIGNORE = f"""
# Kedro Databricks
.databricks
conf/{DEFAULT_TARGET}/**
!conf/{DEFAULT_TARGET}/.gitkeep
""".strip()
"""Content to be added to `.gitignore` for Kedro Databricks configurations."""



import argparse
import logging
from importlib.metadata import version

from kedro_databricks.utils import Command

log = logging.getLogger("kedro_databricks")


def new_project(name):
    """Create a new Kedro project using the Databricks starter."""
    Command(
        ["kedro", "new", "--starter", "databricks-iris", "--name", name], log=log
    ).run()


def run_kedro_databricks_flow(name, destroy=False):
    ver = version("kedro_databricks")
    whl = f"kedro_databricks-{ver}-py3-none-any.whl"
    logging.basicConfig(level=logging.INFO)
    Command(["uv", "build"], log=log).run()
    Command(["mv", f"dist/{whl}", name], log=log).run()
    Command(["uv", "venv"], log=log).run(cwd=f"./{name}")
    Command(["uv", "pip", "install", whl], log=log).run(cwd=f"./{name}")
    Command(
        [
            "uv",
            "run",
            "kedro",
            "databricks",
            "init",
            "--provider",
            "azure",
        ],
        log=log,
    ).run(cwd=f"./{name}")
    Command(
        [
            "uv",
            "run",
            "kedro",
            "databricks",
            "bundle",
            "--overwrite",
        ],
        log=log,
    ).run(cwd=f"./{name}")
    Command(
        [
            "uv",
            "run",
            "kedro",
            "databricks",
            "deploy",
        ],
        log=log,
    ).run(cwd=f"./{name}")
    Command(
        [
            "databricks",
            "bundle",
            "run",
            name.replace("-", "_"),
        ],
        log=log,
    ).run(cwd=f"./{name}")
    if destroy:
        Command(
            [
                "databricks",
                "fs",
                "rm",
                "-r",
                f"dbfs:/FileStore/dev/{name.replace('-', '_')}/",
            ],
            log=log,
        ).run(cwd=f"./{name}")
        Command(["databricks", "bundle", "destroy", "--auto-approve"], log=log).run(
            cwd=f"./{name}"
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run Kedro Databricks flow")
    parser.add_argument(
        "--new",
        action="store_true",
        help="Create a new Kedro Databricks project",
    )
    parser.add_argument(
        "--name",
        type=str,
        default="develop-eggs",
        help="Name of the new Kedro Databricks project",
    )
    parser.add_argument(
        "--destroy",
        action="store_true",
        help="Destroy the Databricks bundle after running",
    )
    args = parser.parse_args()
    if args.new:
        new_project(args.name)
    run_kedro_databricks_flow(args.name, destroy=args.destroy)



from __future__ import annotations

import pytest
from kedro.pipeline import Pipeline, node
from packaging.version import Version

from kedro_databricks.cli.bundle import (
    _load_kedro_env_config,
    _save_bundled_resources,
    bundle,
)
from kedro_databricks.cli.bundle.generate_resources import (
    ResourceGenerator,
    remove_nulls,
    sanitize_name,
    sort_dict,
)
from kedro_databricks.cli.bundle.override_resources import (
    _get_lookup_key,
    _override_workflow,
    _update_list_by_key,
    override_resources,
)
from kedro_databricks.cli.bundle.utils import get_entry_point
from kedro_databricks.constants import OVERRIDE_KEY_MAP
from kedro_databricks.utils import require_databricks_run_script
from tests.utils import WORKFLOW, _generate_task, identity, long_identity, pipeline


def test_bundle(metadata):
    bundle(metadata=metadata, env="fake_env", default_key="default")


def test_generate_workflow(metadata):
    g = ResourceGenerator(metadata, "fake_env")
    assert g._create_workflow("workflow1", pipeline) == WORKFLOW


@pytest.mark.parametrize(
    "node, expected",
    [
        (
            node(
                long_identity,
                ["input", "input1", "input2", "input3", "input4", "input5"],
                ["output", "output1", "output2", "output3", "output4", "output5"],
            ),
            "long_identity_input_input1_input2_input3_input4_input5_output_output1_output2_output3_output4_output",
        ),
        (
            node(identity, ["input"], ["output", "another"]),
            "identity_input_output_another",
        ),
        (node(identity, ["input"], ["output"]), "identity_input_output"),
        (node(identity, ["input"], ["output"], name="a"), "a"),
        (node(identity, ["input"], ["output"], name="a", namespace="abc"), "abc_a"),
        (node(identity, ["input"], ["output"], name="a" * 150), "a" * 100),
    ],
)
def test_sanitize_name(node, expected):
    assert (
        sanitize_name(node) == expected
    ), f"Expected '{expected}', got '{sanitize_name(node)}'"


def test_create_task(metadata):
    g = ResourceGenerator(metadata, "fake_env")
    expected_task = _generate_task("task", ["a", "b"])
    node_a = node(identity, ["input"], ["output"], name="a")
    node_b = node(identity, ["input"], ["output"], name="b")
    assert (
        g._create_task(
            node(
                long_identity,
                ["input", "input1", "input2"],
                ["output", "output1"],
                name="task",
            ),
            [
                node_b,
                node_a,
            ],
        )
        == expected_task
    )


def test_create_task_with_runtime_params(metadata):
    controller = ResourceGenerator(
        metadata, "fake_env", params="key1=value1,key2=value2"
    )
    expected_task = _generate_task(
        "task", ["a", "b"], runtime_params="key1=value1,key2=value2"
    )
    node_a = node(identity, ["input"], ["output"], name="a")
    node_b = node(identity, ["input"], ["output"], name="b")
    assert (
        controller._create_task(
            node(
                long_identity,
                ["input", "input1", "input2"],
                ["output", "output1"],
                name="task",
            ),
            [
                node_b,
                node_a,
            ],
        )
        == expected_task
    )


def test_generate_resources(metadata):
    controller = ResourceGenerator(metadata, "fake_env")
    controller.pipelines = {"__default__": Pipeline([])}
    assert controller.generate_resources(pipeline_name=None) == {}
    controller.pipelines = {
        "__default__": Pipeline([node(identity, ["input"], ["output"], name="node")])
    }
    assert controller.generate_resources(pipeline_name=None) == {
        "fake_project": {
            "resources": {
                "jobs": {
                    "fake_project": {
                        "name": "fake_project",
                        "tasks": [
                            _generate_task("node"),
                        ],
                    },
                },
            },
        },
    }


def test_generate_resources_another_conf(metadata):
    controller = ResourceGenerator(metadata, "fake_env", "sub_conf")
    controller.pipelines = {
        "__default__": Pipeline([node(identity, ["input"], ["output"], name="node")])
    }

    assert controller.generate_resources(pipeline_name=None) == {
        "fake_project": {
            "resources": {
                "jobs": {
                    "fake_project": {
                        "name": "fake_project",
                        "tasks": [
                            _generate_task("node", conf="sub_conf"),
                        ],
                    },
                },
            },
        },
    }


def test_generate_resources_in_a_sorted_manner(metadata):
    controller = ResourceGenerator(metadata, "fake_env")
    controller.pipelines = {
        "__default__": Pipeline(
            [
                node(identity, ["input"], ["b_output"], name="b_node"),
                node(identity, ["input"], ["a_output"], name="a_node"),
            ]
        )
    }
    assert controller.generate_resources(pipeline_name=None) == {
        "fake_project": {
            "resources": {
                "jobs": {
                    "fake_project": {
                        "name": "fake_project",
                        "tasks": [
                            _generate_task("a_node"),
                            _generate_task("b_node"),
                        ],
                    },
                },
            },
        },
    }


def test_generate_resources_for_a_single_pipeline(metadata):
    controller = ResourceGenerator(metadata, "fake_env")
    controller.pipelines = {
        "__default__": Pipeline(
            [
                node(identity, ["input"], ["a_output"], name="a_node"),
            ]
        ),
        "a_pipeline": Pipeline(
            [
                node(identity, ["input"], ["a_output"], name="a_node"),
            ]
        ),
        "b_pipeline": Pipeline(
            [
                node(identity, ["input"], ["b_output"], name="b_node"),
            ]
        ),
    }
    assert controller.generate_resources(pipeline_name="b_pipeline") == {
        "fake_project_b_pipeline": {
            "resources": {
                "jobs": {
                    "fake_project_b_pipeline": {
                        "name": "fake_project_b_pipeline",
                        "tasks": [
                            _generate_task("b_node"),
                        ],
                    },
                },
            },
        },
    }


def test_save_resources(metadata):
    controller = ResourceGenerator(metadata, "fake_env")
    resources = controller.generate_resources()
    overrides = _load_kedro_env_config(metadata, "conf", "fake_env")
    result = {}
    for name, resource in resources.items():
        result[name] = override_resources(resource, overrides, "default")
    _save_bundled_resources(metadata, result, True)
    resource_dir = metadata.project_path.joinpath("resources")
    assert resource_dir.exists(), "Failed to create resources directory"
    assert resource_dir.is_dir(), "resouces is not a directory"

    project_resources = resource_dir.joinpath(f"{metadata.package_name}.yml")
    project_files = ",".join([str(p) for p in resource_dir.iterdir()])
    assert (
        project_resources.exists()
    ), f"Failed to save project resources, {project_files}"


@pytest.mark.parametrize(
    ["key", "expected"],
    [
        ("unknown", None),
        (123, None),
        *[(key, value) for key, value in OVERRIDE_KEY_MAP.items()],
    ],
)
def test_get_lookup_key(key, expected):
    if expected is None:
        with pytest.raises(ValueError):
            _get_lookup_key(key)
        return
    result = _get_lookup_key(key)
    assert result == expected


@pytest.mark.parametrize(
    ["jobs", "overrides", "expected", "error"],
    [
        ({"workflow": "not_dict"}, {}, None, ValueError),
        ({"workflow": {}}, [], None, ValueError),
    ],
)
def test_override_resources(jobs, overrides, expected, error):
    resources = {"resources": {"jobs": jobs}}
    if error:
        with pytest.raises(error):
            override_resources(resources, overrides, "default")
    else:
        result = override_resources(resources, overrides, "default")
        assert result == expected


@pytest.mark.parametrize(
    ["actual", "expected"],
    [
        ("Fake Project", "fake-project"),
        ("Fake Project 123", "fake-project"),
        ("Fake Project 123 456", "fake-project"),
        ("Fake Project #%", "fake-project"),
        ("# Fake Project #%", "fake-project"),
        ("my-project-package", "my-project-package"),
        ("my_project_package", "my_project_package"),
    ],
)
def test_get_entry_point(actual, expected):
    entry_point = get_entry_point(actual)
    assert entry_point == expected, entry_point


@pytest.mark.parametrize(
    ["old", "new", "key", "default", "expected"],
    [
        ([], [], "task_key", {}, []),
        ([], [], "task_key", {"job_cluster_key": "cluster1"}, []),
        (
            [
                {"task_key": "task1", "job_cluster_key": "cluster1"},
                {"task_key": "task2", "job_cluster_key": "cluster2"},
                {"task_key": "task3", "job_cluster_key": "cluster3"},
            ],
            [
                {"task_key": "task1", "job_cluster_key": "cluster4"},
            ],
            "task_key",
            {},
            [
                {"task_key": "task1", "job_cluster_key": "cluster4"},
                {"task_key": "task2", "job_cluster_key": "cluster2"},
                {"task_key": "task3", "job_cluster_key": "cluster3"},
            ],
        ),
        (
            [
                {"task_key": "task1"},
                {"task_key": "task2"},
                {"task_key": "task3"},
            ],
            [
                {"task_key": "task1", "job_cluster_key": "cluster4"},
            ],
            "task_key",
            {"job_cluster_key": "cluster1"},
            [
                {"task_key": "task1", "job_cluster_key": "cluster4"},
                {"task_key": "task2", "job_cluster_key": "cluster1"},
                {"task_key": "task3", "job_cluster_key": "cluster1"},
            ],
        ),
    ],
)
def test_update_list(old, new, key, default, expected):
    result = _update_list_by_key(old, new, key, default)
    assert result == expected, result


@pytest.mark.parametrize(
    ["actual", "order", "expected"],
    [
        (
            {
                "c": 1,
                "a": 2,
                "b": 3,
            },
            ["a", "b", "c"],
            {
                "a": 2,
                "b": 3,
                "c": 1,
            },
        ),
        (
            {
                "a": 1,
                "b": 2,
                "c": 3,
            },
            ["c", "b", "a"],
            {
                "c": 3,
                "b": 2,
                "a": 1,
            },
        ),
    ],
)
def test_sort_dict(actual, order, expected):
    result = sort_dict(actual, order)
    assert result == expected, result


@pytest.mark.parametrize(
    ["value", "expected"],
    [
        (Version("0.19.8"), False),
        (Version("0.19.6"), True),
    ],
)
def test_require_databricks_run_script(value, expected):
    assert require_databricks_run_script(value) == expected, value


@pytest.mark.parametrize(
    ["value", "expected"],
    [
        ([1, 2, 3], [1, 2, 3]),
        ([1, None, 3], [1, 3]),
        ([1, {"a": None}, 3], [1, 3]),
        ([1, [None], 3], [1, 3]),
        ([1, {"a": {"b": None}}, 3], [1, 3]),
        ({}, {}),
        ({"a": 1, "b": None}, {"a": 1}),
        ({"a": 1, "b": {"c": None}}, {"a": 1}),
        ({"a": 1, "b": {"c": {"d": None}}}, {"a": 1}),
    ],
)
def test_remove_nulls_from_dict(value, expected):
    assert remove_nulls(value) == expected


@pytest.mark.parametrize(
    ["dct", "overrides", "expected"],
    [
        (
            {"a": 1, "b": 2},
            {"a": 3},
            {"a": 3, "b": 2},
        ),
        (
            {"a": 1, "b": 2},
            {"c": 3},
            {"a": 1, "b": 2, "c": 3},
        ),
        (
            {"a": 1, "b": 2},
            {"a": 3, "b": 4},
            {"a": 3, "b": 4},
        ),
        (
            {"a": 1, "b": 2},
            {"a": 3, "b": {"c": 4}},
            {"a": 3, "b": {"c": 4}},
        ),
        (
            {"a": 1, "b": 2},
            {"a": 3, "job_clusters": [{"job_cluster_key": "cluster1"}]},
            {"a": 3, "b": 2, "job_clusters": [{"job_cluster_key": "cluster1"}]},
        ),
        (
            {"a": 1, "b": {"c": 2}},
            {
                "a": 3,
                "b": {"c": 3},
                "job_clusters": [{"job_cluster_key": "cluster1"}],
            },
            {"a": 3, "b": {"c": 3}, "job_clusters": [{"job_cluster_key": "cluster1"}]},
        ),
    ],
)
def test_override_workflow(dct, overrides, expected):
    result = _override_workflow(dct, overrides, {}, "default")
    assert result == expected, result


@pytest.mark.parametrize(
    ["args", "error"],
    [
        ([None, {}], TypeError),
        ([{}, None], AttributeError),
        ([{}, None, None], AttributeError),
    ],
)
def test_override_workflow_fail(args, error):
    with pytest.raises(error):
        _override_workflow(*args)



import tomlkit

from kedro_databricks.plugin import commands
from tests import PROJECT_ROOT


def test_version(cli_runner):
    """Test the version command."""
    with open(PROJECT_ROOT / "pyproject.toml", "rb") as f:
        project = tomlkit.load(f)
    result = cli_runner.invoke(commands, ["databricks", "version"])
    expected = project["project"]["version"]  # type: ignore
    assert result.exit_code == 0, (result.exit_code, result.stdout)
    assert (
        expected in result.stdout
    ), f"Expected version {expected} not found in output: {result.stdout}"



from __future__ import annotations

import json
import shutil
from pathlib import Path

import pytest
import yaml

from kedro_databricks.cli.init import (
    _prepare_template,
    _update_gitignore,
    _validate_inputs,
    _write_databricks_run_script,
    init,
)
from kedro_databricks.cli.init.create_target_configs import (
    DatabricksTarget,
    _substitute_file_path,
    create_target_configs,
)
from kedro_databricks.constants import NODE_TYPE_MAP


def test_update_gitignore(metadata):
    _update_gitignore(metadata)
    gitignore_path = metadata.project_path / ".gitignore"
    assert gitignore_path.exists(), "Gitignore not written"
    with open(gitignore_path) as f:
        assert ".databricks" in f.read(), "Databricks not in gitignore"


def test_update_gitignore_does_not_exist(metadata):
    if (metadata.project_path / ".gitignore").exists():
        (metadata.project_path / ".gitignore").unlink()
    _update_gitignore(metadata)
    gitignore_path = metadata.project_path / ".gitignore"
    assert gitignore_path.exists(), "Gitignore not written"
    with open(gitignore_path) as f:
        assert ".databricks" in f.read(), ".databricks not in gitignore"


def test_bundle_init_already_exists(metadata):
    with open(metadata.project_path / "databricks.yml", "w") as f:
        f.write("test")
    with pytest.raises(RuntimeError, match="databricks.yml already exists"):
        init(metadata, "azure", "test")
    with open(metadata.project_path / "databricks.yml") as f:
        assert f.read() == "test", "Databricks config overwritten"


def test_write_databricks_run_script(metadata):
    _write_databricks_run_script(metadata)
    run_script_path = (
        Path(metadata.project_path)
        / "src"
        / metadata.package_name
        / "databricks_run.py"
    )
    assert run_script_path.exists(), "Databricks run script not written"


@pytest.mark.parametrize(
    ["actual", "expected"],
    [
        (
            "file_path: /dbfs/FileStore/develop_eggs/data/01_raw/file.csv",
            "file_path: ${_file_path}/data/01_raw/file.csv",
        ),
        (
            "file_path: /dbfs/develop_eggs/data/01_raw/file.csv",
            "file_path: ${_file_path}/data/01_raw/file.csv",
        ),
        (
            "file_path: /dbfs/FileStore/develop_eggs/data/01_raw/file.csv",
            "file_path: ${_file_path}/data/01_raw/file.csv",
        ),
        (
            "file_path: data/0_raw/file.csv",
            "file_path: ${_file_path}/data/0_raw/file.csv",
        ),
        (
            "file_path: data/012_raw/file.csv",
            "file_path: ${_file_path}/data/012_raw/file.csv",
        ),
        (
            "file_path: /custom/path/data/01_raw/file.csv",
            "file_path: ${_file_path}/data/01_raw/file.csv",
        ),
        ("data/01_raw/file.csv", "data/01_raw/file.csv"),
    ],
)
def test_substitute_file_path(actual, expected):
    result = _substitute_file_path(actual)
    assert result == expected, f"\n{result}\n{expected}"


def test_create_target_configs(metadata, monkeypatch):
    monkeypatch.setattr(
        DatabricksTarget,
        "_get_metadata",
        lambda *args, **kwargs: {
            "workspace": {
                "file_path": "/Workspace/<your-volume-name>/develop_eggs/data/01_raw/file.csv",
            }
        },  # noqa: E501
    )
    with open(metadata.project_path / "databricks.yml", "w") as f:
        yaml.safe_dump(
            {
                "bundle": {
                    "name": "develop_eggs",
                },
                "targets": {
                    "dev": {
                        "mode": "development",
                        "workspace": {
                            "host": "https://<your-volume-name>.databricks.com",
                        },
                    }
                },
            },
            f,
        )
    create_target_configs(
        metadata, NODE_TYPE_MAP.get("azure", ""), "test", single_user_default=False
    )


def test_validate_inputs_unknown_provider(metadata):
    with pytest.raises(ValueError, match="Invalid provider 'unknown'"):
        _validate_inputs(metadata, "unknown")


def test_validate_inputs(metadata):
    if (metadata.project_path / "databricks.yml").exists():
        (metadata.project_path / "databricks.yml").unlink()
    config_path, node_type_id = _validate_inputs(metadata, "azure")
    assert config_path == metadata.project_path / "databricks.yml"
    assert node_type_id == NODE_TYPE_MAP["azure"]


def test_validate_inputs_config_already_exists(metadata):
    if not (metadata.project_path / "databricks.yml").exists():
        (metadata.project_path / "databricks.yml").write_text("test")
    with pytest.raises(RuntimeError, match="databricks.yml already exists"):
        _validate_inputs(metadata, "azure")


def test_prepare_template(metadata):
    assets_dir, template_params = _prepare_template(metadata)
    assert assets_dir.exists(), "Assets directory not created"
    params = template_params.read_text()
    try:
        conf = json.loads(params)
        assert conf["project_name"] == metadata.package_name, "Project name not set"
        assert conf["project_slug"] == metadata.package_name, "Package name not set"
    except Exception as e:
        raise ValueError(f"Failed to load template params: {e} - {params}")
    shutil.rmtree(assets_dir)



import json
import shutil
import tempfile
from pathlib import Path

import tomlkit
from kedro.framework.startup import ProjectMetadata

from kedro_databricks.cli.init.create_target_configs import create_target_configs
from kedro_databricks.constants import GITIGNORE, NODE_TYPE_MAP, TEMPLATES
from kedro_databricks.logger import get_logger
from kedro_databricks.utils import (
    Command,
    assert_databricks_cli,
    require_databricks_run_script,
)

log = get_logger("init")


def init(
    metadata: ProjectMetadata, provider: str, default_key: str, *databricks_args: str
):
    """Initialize a Databricks Asset Bundle.

    This function creates a Databricks Asset Bundle in the current project
    directory. It also creates a Databricks configuration file and a
    Databricks target configuration file.

    Args:
        metadata (ProjectMetadata): The project metadata.
        provider (str): The provider to use. Valid providers are "azure", "aws", or "gcp".
        default_key (str): The default key to use for the Databricks target.
        *databricks_args: Additional arguments to be passed to the `databricks` CLI.

    Raises:
        RuntimeError: If the `databricks` CLI is not installed or the wrong version is used.
        ValueError: If the provider is not valid.
    """
    assert_databricks_cli()
    log.info("Initializing Databricks Asset Bundle...")
    config_path, node_type_id = _validate_inputs(metadata, provider)
    _databricks_init(metadata, *databricks_args)
    log.info(f"Created {config_path.relative_to(metadata.project_path)}")
    create_target_configs(metadata, node_type_id=node_type_id, default_key=default_key)
    _update_gitignore(metadata)
    if require_databricks_run_script():  # pragma: no cover - Might be removed in future
        log.warning(
            "Kedro version less than 0.19.8 requires a script to run tasks on Databricks. "
        )
        _write_databricks_run_script(metadata)
    log.info(
        f"Successfully initialized Databricks Asset Bundle in {metadata.project_path}"
    )


def _validate_inputs(metadata: ProjectMetadata, provider: str):
    if provider not in NODE_TYPE_MAP:
        raise ValueError(
            f"Invalid provider '{provider}'. Valid providers are: {', '.join(NODE_TYPE_MAP.keys())}"
        )
    config_path = metadata.project_path / "databricks.yml"
    if config_path.exists():
        raise RuntimeError(
            f"{config_path.relative_to(metadata.project_path)} already exists."
        )
    return config_path, NODE_TYPE_MAP[provider]


def _databricks_init(metadata: ProjectMetadata, *databricks_args):
    assets_dir, template_params = _prepare_template(metadata)
    init_cmd = [
        "databricks",
        "bundle",
        "init",
        assets_dir.as_posix(),
        "--config-file",
        template_params.as_posix(),
        "--output-dir",
        metadata.project_path.as_posix(),
    ] + list(databricks_args)
    result = Command(init_cmd, log=log, warn=True).run()
    if result.returncode != 0:  # pragma: no cover
        err = "\n".join(result.stdout)
        raise RuntimeError(f"Failed to initialize Databricks Asset Bundle\n{err}")
    shutil.rmtree(assets_dir)


def _prepare_template(metadata: ProjectMetadata):
    assets_dir = Path(tempfile.mkdtemp())
    shutil.copy(
        str(TEMPLATES / "databricks_template_schema.json"),
        str(assets_dir / "databricks_template_schema.json"),
    )
    template_dir = assets_dir / "template"
    template_dir.mkdir(exist_ok=True)
    shutil.copy(
        str(TEMPLATES / "databricks.yml.tmpl"),
        str(template_dir / "databricks.yml.tmpl"),
    )
    config = {
        "project_name": metadata.package_name,
        "project_slug": metadata.package_name,
    }
    params_file = assets_dir / "params.json"
    params_file.touch()
    params_file.write_text(json.dumps(config))
    return assets_dir, params_file


def _update_gitignore(metadata: ProjectMetadata):
    gitignore_path = metadata.project_path / ".gitignore"
    if not gitignore_path.exists():
        gitignore_path.touch()
    current_gitignore = gitignore_path.read_text()
    with open(gitignore_path, "w") as f:
        f.write(f"{GITIGNORE}\n{current_gitignore}")


def _write_databricks_run_script(metadata: ProjectMetadata):
    script_path = (
        metadata.project_path / "src" / metadata.package_name / "databricks_run.py"
    )
    toml_path = metadata.project_path / "pyproject.toml"
    shutil.copy(str(TEMPLATES / "databricks_run.py"), str(script_path))
    log.info(f"Wrote {script_path.relative_to(metadata.project_path)}")

    with open(toml_path) as f:
        toml = tomlkit.load(f)

    scripts = toml.get("project", {}).get("scripts", {})
    if "databricks_run" not in scripts:
        scripts["databricks_run"] = f"{metadata.package_name}.databricks_run:main"
        toml["project"]["scripts"] = scripts  # type: ignore

    log.info(f"Added script to {toml_path.relative_to(metadata.project_path)}")
    with open(toml_path, "w") as f:
        tomlkit.dump(toml, f)



from pathlib import Path

import mkdocs_gen_files

mkdocs_gen_files.log.info("Generating example pages...")
root = Path(__file__).parent.parent
examples_dir = root / "examples"


def gen():
    """Generate the code reference pages."""
    for p in sorted(examples_dir.iterdir()):
        if not p.is_dir() or p.name.startswith("."):
            continue

        example = _load_example(p)

        with mkdocs_gen_files.open(f"examples/{p.name}.md", "w") as f:
            f.write("\n".join(example) + "\n")
    pass


def _load_data(p: Path):
    description = ""
    if (p / "README.md").exists():
        description = (p / "README.md").read_text().strip()
    databricks_config = _parse_file(p / "databricks.yml")
    resources_config = _parse_file(p / "resources.yml")
    result_config = _parse_file(p / "result.yml")
    return databricks_config, resources_config, result_config, description


def _load_example(p: Path) -> list[str]:
    """Load an example from the given path."""
    databricks_config, resources_config, result_config, description = _load_data(p)
    result_highlight = _get_hl_lines(resources_config, result_config)
    parts = []
    parts.append(description)
    parts.append("")
    parts.append('=== "conf/[env]/databricks.yml"')
    parts.append("    ```yaml")
    parts.extend(_add_padding(databricks_config, leftpad=4))
    parts.append("    ```")
    parts.append("")
    parts.append('=== "Before: resources/<pipeline>.yml"')
    parts.append("    ```yaml")
    parts.extend(_add_padding(resources_config, leftpad=4))
    parts.append("    ```")
    parts.append("")
    parts.append('=== "After: resources/<pipeline>.yml"')
    parts.append(f'    ```yaml hl_lines="{result_highlight}"')
    parts.extend(_add_padding(result_config, leftpad=4))
    parts.append("    ```")
    return parts


def _get_hl_lines(old, new):
    old = old.copy()
    new = new.copy()
    i, j = 0, 0
    hl_lines = []
    while j < len(new):
        i = min(i, len(old) - 1)
        if old[i] != new[j]:
            # +1 for 1-based indexing in markdown hl_lines
            hl_lines.append(j + 1)
            j += 1
        elif old[i] == new[j]:
            i += 1
            j += 1
        else:
            raise ValueError("Unexpected case in diffing")
    return " ".join(str(x) for x in hl_lines)


def _parse_file(file_path: Path) -> list[str]:
    """Parse a file and return its content."""
    return [line for line in file_path.read_text().strip().splitlines()]


def _add_padding(lines: list[str], leftpad: int = 4) -> list[str]:
    """Add padding to each line in the list."""
    padding = " " * leftpad
    return [padding + line for line in lines]


gen()
